@@ -14,6 +14,6 @@ void hello() {
 EOF
 
 ! cc --ld-path=./ld64 -shared -o $t/b.dylib $t/a.o -Wl,-pagezero_size,0x1000 >& $t/log
-fgrep -q ' -pagezero_size option can only be used when linking a main executable' $t/log
+grep -Fq ' -pagezero_size option can only be used when linking a main executable' $t/log
 
 echo OK
