@@ -1,15 +1,9 @@
 #!/bin/bash
 export LC_ALL=C
 set -e
-CC="${TEST_CC:-cc}"
-CXX="${TEST_CXX:-c++}"
-GCC="${TEST_GCC:-gcc}"
-GXX="${TEST_GXX:-g++}"
-OBJDUMP="${OBJDUMP:-objdump}"
-MACHINE="${MACHINE:-$(uname -m)}"
 testname=$(basename "$0" .sh)
 echo -n "Testing $testname ... "
-t=out/test/macho/$MACHINE/$testname
+t=out/test/macho/$(uname -m)/$testname
 mkdir -p $t
 
 cat > $t/libfoo.tbd <<'EOF'
@@ -30,17 +24,17 @@ exports:
 ...
 EOF
 
-cat <<EOF | clang -o $t/a.o -c -xc -
+cat <<EOF | cc -o $t/a.o -c -xc -
 void foo();
 int main() { foo(); }
 EOF
 
-clang --ld-path=./ld64 -o $t/exe1 $t/libfoo.tbd $t/a.o \
+cc --ld-path=./ld64 -o $t/exe1 $t/libfoo.tbd $t/a.o \
   -Wl,-platform_version,macos,20.0,20.0 >& /dev/null
 
 otool -L $t/exe1 | grep -q /foo
 
-clang --ld-path=./ld64 -o $t/exe2 $t/libfoo.tbd $t/a.o \
+cc --ld-path=./ld64 -o $t/exe2 $t/libfoo.tbd $t/a.o \
   -Wl,-platform_version,macos,25.0,21.0 >& /dev/null
 
 otool -L $t/exe2 | grep -q /bar
